unit DUTUtilsGraphics;
{..$DEFINE DEBUG}
(*
   Daniel U. Thibault
   <D.U.Thibault@Bigfoot.com>
   31 January 2002
    1998-2002 Daniel U. Thibault.  All Rights Reserved.

   This is freeware.  No legally binding promises.  If you use
   any part of this code, you must include an acknowledgement in the
   finished code or executable (in the About box and/or the Help file).

   Graphics utilities.
*)
interface

uses
   Windows,  { HPalette, MAX_PATH, etc. }
   SysUtils, { TFileName, PByteArray, ANSIPos, etc. }
   Graphics; { TCanvas, TBitmap, TIcon }

{ #################### Graphic Utilities #################### }

function WMF_to_EMF(WMF : HMetaFile; DC : HDC; Width, Height, Inch : Integer) : HEnhMetaFile;
{
   Converts a WMF (old-format Windows Metafile) into an EMF (Enhanced Metafile).
   WMF is the WMF's handle (returned by most Windows API methods);
   DC is the device context (or information context).  Zero = current output device (the screen).
   Width and Height are a suggested size in pixels.  Zero = let the system figure it out.
   Inch is the length of an inch in metafile units, when known.
      Zero = ignore; this will force Width and Height to be ignored as well.
}

function DuplicateBitmap(var SourceGraphic : TGraphic; Monochrome : Boolean) : Windows.HBITMAP;
{
   Creates a duplicate HBITMAP object from the specified TGraphic
   object (either a TBitmap or a TIcon --and their descendants).
   Monochrome indicates whether the duplicate should be monochrome or
   retain the source's colour depth.
}

function GetPaletteNumEntries(Palette : HPalette) : Word;
{
   Returns the number of palette entries in the argument palette.
}

function GetSystemPaletteNumEntries(DC : HDC) : Word;
{
   Returns the number of palette entries in the system palette for the specified device context.
}

function MinimizeString(const Text : String; Canvas : TCanvas; MaxLen : Integer) : String;
{
   Returns an elided Text such that it'll fit within MaxLen pixels in the Canvas.
   (i.e. Unmodified Text OR Text[0..x] plus an ellipsis, where x < Length(Text))
   It is modelled after MinimizeName.
}

function MinimizeFileName(const FileName : TFileName; Canvas : TCanvas; MaxLen : Integer) : TFileName;
{
   Minimizes the FileName so it'll fit in Width pixels on the Canvas.
   Assumes FileName is fully qualified.
   Replacement for FileCtrl.pas' MinimizeName function that handles network paths correctly.
   Also now returns a null string if the file title + extension won't fit.
   This will be, in order of feasability:
   [Drive][Path][Title][Extension]
   [Drive]...[foreshortened Path][Title][Extension] (Path is foreshortened by removing whole subdirectories)
   [Drive]...\[Title][Extension]
   ...\[Title][Extension]
   [Title][Extension]
   [Null string]
}

implementation

uses
{$IFDEF DEBUG}
   DLLDebug,
{$ENDIF}
(* System,   { AssignFile, Pos, Copy, Length } *)
   Registry, { TRegistry }
   Consts,   { SInvalidDate, sInvalidMetafile }
   ShlObj,   { SHGetSpecialFolderLocation, etc. }
   ShellAPI, { TShellExecuteInfo }
   TypInfo,  { PTypeInfo, etc. }
   DUTUtilsMisc; { FlagIncluded }

{ #################### Graphic Utilities #################### }

function WMF_to_EMF(WMF : HMetaFile; DC : HDC; Width, Height, Inch : Integer) : HEnhMetaFile;
{
   Converts a WMF (old-format Windows Metafile) into an EMF (Enhanced Metafile).
   WMF is the WMF's handle (returned by most Windows API methods);
   DC is the device context (or information context).  Zero = current output device (the screen).
   Width and Height are a suggested size in pixels.  Zero = let the system figure it out.
   Inch is the length of an inch in metafile units, when known.
      Zero = ignore; this will force Width and Height to be ignored as well.
}
var
   cbBuffer  : Integer;
   lpbBuffer : String;
   WMFPict   : TMetaFilePict;
begin
   //Obtain the required buffer size
   cbBuffer := GetMetaFileBitsEx(WMF, 0, Nil);
   if cbBuffer = 0 then raise Exception.CreateFmt('Corrupt WMF - GetMetaFileBitsEx failed to evaluate buffer size: <<%s>>', [SysErrorMessage(GetLastError)]);
   //Create the buffer
   lpbBuffer := StringOfChar(' ', cbBuffer);
   //Read the WMF into the buffer
   cbBuffer := GetMetaFileBitsEx(WMF, cbBuffer, PChar(lpbBuffer));
   if cbBuffer = 0 then RaiseLastWin32Error;
   //cbBuffer now contains the number of bytes copied to the buffer
   //Convert the buffer contents to EMF
   //This part is similar to what TMetafile.ReadWMFStream does with its BitMem buffer
   with WMFPict do begin
      mm := MM_ANISOTROPIC;
      //MM_HIMETRIC units (0,01 mm) (1 in = 25,4 mm = 2540 hmu)
      xExt := 0; yExt := 0;
      if Inch <> 0 then begin
         xExt := MulDiv(2540, Width, Inch);
         yExt := MulDiv(2540, Height, Inch);
      end else if (0 <> GetDeviceCaps(DC, LOGPIXELSX)) and (0 <> GetDeviceCaps(DC, LOGPIXELSY)) then begin
//    end else begin
         xExt := MulDiv(2540, Width, GetDeviceCaps(DC, LOGPIXELSX));
         yExt := MulDiv(2540, Height, GetDeviceCaps(DC, LOGPIXELSY));
//       xExt := MulDiv(2540, Width, 96);
//       yExt := MulDiv(2540, Height, 96);
      end; { if }
      hMF := WMF; //Ignored
   end; { with }
   //A null hDC indicates the current output device
   Result := SetWinMetaFileBits(cbBuffer, PChar(lpbBuffer), DC, WMFPict);
   if Result <> 0 then exit;
   raise EInvalidGraphic.Create(sInvalidMetafile);
end; { WMF_to_EMF }

function DuplicateBitmap(var SourceGraphic : TGraphic; Monochrome : Boolean) : Windows.HBITMAP;
{
   Creates a duplicate HBITMAP object from the specified TGraphic
   object (either a TBitmap or a TIcon --and their descendants).
   Monochrome indicates whether the duplicate should be monochrome or
   retain the source's colour depth.
}
var
   vTB     : Windows.TBitmap;
   PictSiz : Integer;
   NewPict : PByteArray;
   NewBmp  : Graphics.TBitmap;
begin
   //Return a null hBitmap in case of failure
   Result := 0;
   if not Assigned(SourceGraphic) then Exit;

   if (SourceGraphic is Graphics.TBitmap) then begin
      if (not Monochrome) or ((SourceGraphic as Graphics.TBitmap).Monochrome) then begin
         GetObject((SourceGraphic as Graphics.TBitmap).Handle, SizeOf(vTB), @vTB);
         PictSiz := (SourceGraphic.Height * SourceGraphic.Width * vTB.bmBitsPixel * vTB.bmPlanes) div 8;
         GetMem(NewPict, PictSiz);
         GetBitmapBits((SourceGraphic as Graphics.TBitmap).Handle, PictSiz, NewPict);
         Result := CreateBitmap(SourceGraphic.Width, SourceGraphic.Height, vTB.bmPlanes, vTB.bmBitsPixel, NewPict);
         Exit;
      end; { then }
   end else if not (SourceGraphic is Graphics.TIcon) then begin
      Exit;
   end; { else }
   //Must reduce colour depth OR SourceGraphic is TIcon
   NewBmp := Graphics.TBitmap.Create;
   try
      NewBmp.Width := SourceGraphic.Width;
      NewBmp.Height := SourceGraphic.Height;
      NewBmp.Monochrome := Monochrome;
      NewBmp.Canvas.Draw(0, 0, SourceGraphic);
      Result := DuplicateBitmap(TGraphic(NewBmp), Monochrome);
   finally
      NewBmp.Free;
   end; { try }
end; { DuplicateBitmap }

function GPE(hPal : HPalette; StartIndex, NumEntries : Word; pPaletteEntries : Pointer) : Word; stdcall;
   external gdi32 name 'GetPaletteEntries';
function GetPaletteNumEntries(Palette : HPalette) : Word;
{
   Returns the number of palette entries in the argument palette.
}
begin
   Result := GPE(Palette, 0, 0, Nil);
end; { GetPaletteNumEntries }

function GSPE(DC : HDC; StartIndex, NumEntries : Word; pPaletteEntries : Pointer) : Word; stdcall;
   external gdi32 name 'GetSystemPaletteEntries';
function GetSystemPaletteNumEntries(DC : HDC) : Word;
{
   Returns the number of palette entries in the system palette for the specified device context.
}
begin
   Result := GSPE(DC, 0, 0, Nil);
end; { GetSystemPaletteNumEntries }

function MinimizeString(const Text : String; Canvas : TCanvas; MaxLen : Integer) : String;
{
   Returns an elided Text such that it'll fit within MaxLen pixels in the Canvas.
   (i.e. Unmodified Text OR Text[0..x] plus an ellipsis, where x < Length(Text))
   It is modelled after MinimizeName.
}
var
   i : Integer;
begin
   Result := Text;
   if Canvas.TextWidth(Result) <= MaxLen then exit;
   i := Length(Text);
   { Must handle case where Text is very large }
   while (i > 0) and (Canvas.TextWidth(Result) > MaxLen) do begin
      { We moderate a Newton-Raphson approach so we don't shrink Result too fast }
      i := (i + ((MaxLen*i) div Canvas.TextWidth(Result))) div 2;
      //Can we get stuck in an infinite loop here?  I don't think so.
      //Since Canvas.TextWidth > MaxLen, this guarantees i will shrink by at least one.
      Result := Copy(Text, 1, i) + '...';
   end; { while }
   { May have shortened i too much }
   while Canvas.TextWidth(Copy(Text, 1, i + 1) + '...') <= MaxLen do begin
      Inc(i);
   end; { while }
   Result := Copy(Text, 1, i) + '...';
   { Maybe even '...' is too long }
   if Canvas.TextWidth(Result) > MaxLen then Result := '';
end; { MinimizeString }

function MinimizeFileName(const FileName : TFileName; Canvas : TCanvas; MaxLen : Integer) : TFileName;
{
   Minimizes the FileName so it'll fit in Width pixels on the Canvas.
   Assumes FileName is fully qualified.
   Replacement for FileCtrl.pas' MinimizeName function that handles network paths correctly.
   Also now returns a null string if the file title + extension won't fit.
   This will be, in order of feasability:
   [Drive][Path][Title][Extension]
   [Drive]...[foreshortened Path][Title][Extension] (Path is foreshortened by removing whole subdirectories)
   [Drive]...\[Title][Extension]
   ...\[Title][Extension]
   [Title][Extension]
   [Null string]
}
   procedure CutFirstDirectory(var S : TFileName);
   {
      Removes the first subdirectory in the path S.
      S is assumed to end with '\'; the beginning '\' is optional.
      '[\]path1\path2\'     becomes '[\]...\path2\'
      '[\]...\path1\path2\' becomes '[\]...\path2\'
      '[\]...\path\'        becomes '[\]...\'
      '[\]...\'             becomes '\'
      '\'                   becomes ''
   }
   var
      Root : Boolean;
      P    : Integer;
   begin
      if S = '\' then
         S := ''
      else begin
         if S[1] = '\' then begin
            Root := True;
            Delete(S, 1, 1);
         end else
            Root := False;
         if S[1] = '.' then
            { Assumes '...\' }
            Delete(S, 1, 4);
         P := ANSIPos('\', S);
         if P <> 0 then begin
            Delete(S, 1, P);
            S := '...\' + S;
         end else
            S := '';
         if Root then
            S := '\' + S;
      end; { else }
   end; { CutFirstDirectory }
var
   Drive,
   Dir,
   Name  : TFileName;
begin
   Result := FileName;
   Dir := ExtractFilePath(Result);
   Name := ExtractFileName(Result);

   if (Length(Dir) >= 2) and (Dir[2] = ':') then begin
      Drive := Copy(Dir, 1, 2);
      Delete(Dir, 1, 2);
   //ADDITIONAL CODE STARTS HERE
   end else if (Length(Dir) >= 2) and (Dir[2] = '\') then begin
      { UNC Drive! }
      Drive := Copy(Dir, 1, 2 + Pos('\', Copy(Dir, 3, Length(Dir))));
      Delete(Dir, 1, Length(Drive));
   //ADDITIONAL CODE ENDS HERE
   end else
      Drive := '';
   while ((Dir <> '') or (Drive <> '')) and (Canvas.TextWidth(Result) > MaxLen) do begin
      if Dir = '\...\' then begin
         Drive := '';
         Dir := '...\';
      end else if Dir = '' then
         Drive := ''
      else
         CutFirstDirectory(Dir);
      Result := Drive + Dir + Name;
   end; { while }
   //ADDITIONAL CODE BEGINS HERE
   if Canvas.TextWidth(Result) > MaxLen then Result := '';
   //ADDITIONAL CODE ENDS HERE
end; { MinimizeFileName }

end. { DUTUtilsGraphics }
