unit DUTUtilsStrings;
(*
   Daniel U. Thibault
   <D.U.Thibault@Bigfoot.com>
   21 July 1999
    1998-1999 Daniel U. Thibault.  All Rights Reserved.

   This is freeware.  No legally binding promises.  If you use
   any part of this code, you must include an acknowledgement in the
   finished code or executable (in the About box and/or the Help file).

   String manipulation utilities.
*)
interface

{ #################### String Manipulation #################### }

function WideCharLen(Source : PWideChar) : Integer;
{
   Returns a null-terminated wide string's length,
   in double-byte characters (excluding the null).
}

function DUTStringToWideChar(const Source : string; Dest : PWideChar; DestSize : Integer) : PWideChar;
{
   Fixes a number of problems with System.pas' StringToWideChar.
   This replacement function raises an exception if things go awry.
}

function ANSItoASCII(const S : String) : String;
{
   Converts an ANSI string to an ASCII string.
   Characters are replaced by stripping the accents out, diphtongs are expanded,
   some other characters are replaced by close relatives (e.g. per-thousand becomes percent).
   Untranslatable characters are changed to $7F.
   Not locale-sensitive and certainly pretty crude.
}

function Repl(const OldStr : String; const NewStr : String; const S : String) : String;
{
   Replaces all instances of OldStr in S by NewStr and returns the result.
   This is recursive, so if you replace '  ' (double-blank) by ' ' (single-blank),
   all multiple blanks will dwindle to a single blank.
   The reverse isn't recommended as it'll create an infinite loop!
}

function MultiRepl(const OldStrs : Array of String; const NewStrs : Array of String; const S : String) : String;
{
   Replaces all instances of OldStrs[0..] in S by NewStrs[0..] and returns the result.
   MultiRepl first calls Repl for the OldStrs[0], NewStrs[0] pair and then works its way to higher pairs.
   If OldStrs and NewStrs don't have the same number of elements, an exception is raised.
}

function DUTStrToDate(const S: string): TDateTime;
{
   SysUtils' StrToDate fails miserably because it can't handle
   the ddd, mmm and mmmm format specifiers in ShortDateFormat.
   This routine handles m, mm, mmm, mmmm, d, dd, ddd, yy and yyyy.
}

function IntToRoman(Num : integer) : string;
{
   Returns the roman numeral representation of Num.
An exception is raised if Num is outside the range [1..3999].
   Larger numbers are possible but cannot be represented
in ANSI characters; 10,000, for example, is represented by an
X surmounted by a horizontal bar --e.g. "overlined X".
   An improved and corrected version of Shane Holmes' romanNum function.
}

implementation

uses
   Windows,  { SetLastError, MultiByteToWideChar, etc. }
   SysUtils; { Exception, SysErrorMessage, DecodeDate, etc. }

const
   SInvalidDateFormat = '''''%s'''' is not a valid date format string';

{ #################### String Manipulation #################### }

function WideCharLen(Source : PWideChar) : Integer;
{
   Returns a null-terminated wide string's length,
   in double-byte characters (excluding the null).
}
begin
   Result := 0;
   while Source[Result] <> #0 do Inc(Result);
end; { WideCharLen }

function DUTStringToWideChar(const Source : string; Dest : PWideChar; DestSize : Integer) : PWideChar;
{
   Fixes a number of problems with System.pas' StringToWideChar.
   This replacement function raises an exception if things go awry.
}
const
   MB_ERR_INVALID_CHARS = $00000008; //Missing from Windows.pas
var
   Code : Integer;
begin
   SetLastError(NO_ERROR);
   Code := MultiByteToWideChar(CP_ACP { ANSI code page }, MB_ERR_INVALID_CHARS, PChar(Source), Length(Source), Dest, DestSize - 1);
   if Code = 0 then
      raise Exception.CreateFmt('DUTStringToWideChar: "%s" (Error code %d)', [SysErrorMessage(GetLastError), GetLastError]);
   Dest[Code] := #0;
   Result := Dest;
end; { DUTStringToWideChar }

function ANSItoASCII(const S : String) : String;
{
   Converts an ANSI string to an ASCII string.
   Characters are replaced by stripping the accents out, diphtongs are expanded,
   some other characters are replaced by close relatives (e.g. per-thousand becomes percent).
   Untranslatable characters are changed to $7F.
   Not locale-sensitive and certainly pretty crude.
}
var
   i : Integer;
begin
   Result := MultiRepl(['', '', '', '', '', '', '',  '', '', '', '', '',
                        '', '', '', '', '', '', '', '', '', '', '', '',
                        '', '', '', '', '', '', '',
                        '', '', '', '', '', '', '',  '', '', '', '', '',
                        '', '', '', '', '', '', '', '', '', '', '', '',
                        '', '', '', '', '', '', '',
                        #160, '', '', '', '',   '',
                        '', '', '', '', '', '', '', '', '',  '',
                        '', '', '', '', '', '', '', '',  '',   '',   '', '',
                        '', '', '', '', '', '', '', '', '',  '', '', '', '',   '',   '',   '', ''],
                       ['a', 'a', 'a', 'a', 'a', 'a', 'ae', 'c', 'e', 'e', 'e', 'e',
                        'i', 'i', 'i', 'i', 'n', 'o', 'o', 'o', 'o', 'o', 'o', 'oe',
                        's', 'u', 'u', 'u', 'u', 'y', 'y',
                        'A', 'A', 'A', 'A', 'A', 'A', 'AE', 'C', 'E', 'E', 'E', 'E',
                        'I', 'I', 'I', 'I', 'N', 'O', 'O', 'O', 'O', 'O', 'O', 'OE',
                        'S', 'U', 'U', 'U', 'U', 'Y', 'Y',
                        ' ',  '_', '%', 'f', '...', 'ss',
                        '"', '"', '"', #27, #27, #27, '<', '>', '<<', '>>',
                        #27, '^', '~', '"', '-', '-', '-', 'tm', '(c)', '(r)', '!', '?',
                        'c', 'L', '$', '$', 'Y', '|', 'a', 'o', '+-', '1', '2', '3', '1/4', '1/2', '3/4', '*', '/'],
                       S);
   for i := 1 to Length(Result) do begin
      if Ord(Result[i]) > 127 then Result[i] := #127;
   end; { for }
end; { ANSItoASCII }

function Repl(const OldStr : String; const NewStr : String; const S : String) : String;
{
   Replaces all instances of OldStr in S by NewStr and returns the result.
   This is recursive, so if you replace '  ' (double-blank) by ' ' (single-blank),
   all multiple blanks will dwindle to a single blank.
   The reverse isn't recommended as it'll create an infinite loop!
}
begin
   Result := S;
   while Pos(OldStr, Result) > 0 do begin
      Result := Copy(Result, 1, Pos(OldStr, Result) - 1) + NewStr + Copy(Result, Pos(OldStr, Result) + Length(OldStr), Length(Result));
   end; { while }
end; { Repl }

function MultiRepl(const OldStrs : Array of String; const NewStrs : Array of String; const S : String) : String;
{
   Replaces all instances of OldStrs[0..] in S by NewStrs[0..] and returns the result.
   MultiRepl first calls Repl for the OldStrs[0], NewStrs[0] pair and then works its way to higher pairs.
   If OldStrs and NewStrs don't have the same number of elements, an exception is raised.
}
var
   i : Integer;
begin
   if High(OldStrs) <> High(NewStrs) then raise Exception.Create('MultiRepl: Mismatched string array sizes.');
   Result := S;
   for i := 0 to High(OldStrs) do begin
      Result := Repl(OldStrs[i], NewStrs[i], Result);
   end; { for }
end; { MultiRepl }

function DUTStrToDate(const S: string): TDateTime;
{
   SysUtils' StrToDate fails miserably because it can't handle
   the ddd, dddd, mmm and mmmm format specifiers in ShortDateFormat.
   This routine handles m, mm, mmm, mmmm, d, dd, ddd, dddd, yy and yyyy.
   S is assumed to follow the ShortDateFormat.
}
var
   i, j, ii, jj,
   M                : Integer;
   Year, Month, Day,
   CurrentYear      : Word;
   SS               : String;
begin
   //Obtain CurrentYear
   DecodeDate(Date, CurrentYear, Month, Day);
   Year := 0; Month := 0; Day := 0;
   i := 1; ii := 1;
   while i <= Length(ShortDateFormat) do begin
      if Length(S) < ii then
         raise EConvertError.CreateFmt(SInvalidDate, [S]);
      //Find next DateSeparator
      j := (i - 1) + Pos(DateSeparator, Copy(ShortDateFormat + DateSeparator, i, 2 + Length(ShortDateFormat) - i));
      SS := Copy(ShortDateFormat, i, j - i);
      if          0 = CompareText(SS, 'yy') then begin
         jj := ii + 2;
         //Year is S[ii..jj[
         if Length(S) < (jj - 1) then
            raise EConvertError.CreateFmt(SInvalidDate, [S]);
         try
            Year := StrToInt(Copy(S, ii, jj - ii));
            //Add current century
            Inc(Year, 100 * CurrentYear div 100);
            //Sliding window:
            //If year is more than 50 years in the past,
            //assume it really is in the future
            if (CurrentYear - Year) > 50 then Inc(Year, 100);
         except
            raise EConvertError.CreateFmt(SInvalidDate, [S]);
         end; { try }
      end else if 0 = CompareText(SS, 'yyyy') then begin
         jj := ii + 4;
         //Year is S[ii..jj[
         if Length(S) < (jj - 1) then
            raise EConvertError.CreateFmt(SInvalidDate, [S]);
         try
            Year := StrToInt(Copy(S, ii, jj - ii));
         except
            raise EConvertError.CreateFmt(SInvalidDate, [S]);
         end; { try }
      end else if 0 = CompareText(SS, 'm') then begin
         jj := (ii - 1) + Pos(DateSeparator, Copy(S + DateSeparator, ii, 2 + Length(S) - ii));
         //Month is S[ii..jj[
         try
            Month := StrToInt(Copy(S, ii, jj - ii));
         except
            raise EConvertError.CreateFmt(SInvalidDate, [S]);
         end; { try }
      end else if 0 = CompareText(SS, 'mm') then begin
         jj := ii + 2;
         //Month is S[ii..jj[
         try
            Month := StrToInt(Copy(S, ii, jj - ii));
         except
            raise EConvertError.CreateFmt(SInvalidDate, [S]);
         end; { try }
      end else if 0 = CompareText(SS, 'mmm') then begin
         jj := (ii - 1) + Pos(DateSeparator, Copy(S + DateSeparator, ii, 2 + Length(S) - ii));
         //Month is S[ii..jj[ as a ShortMonthNames string
         for M := 1 to 12 do begin
            if 0 = CompareText(Copy(S, ii, jj - ii), ShortMonthNames[M]) then begin
               Month := M;
               break;
            end; { then }
         end; { for }
         if Month = 0 then
            raise EConvertError.CreateFmt(SInvalidDate, [S]);
      end else if 0 = CompareText(SS, 'mmmmm') then begin
         jj := (ii - 1) + Pos(DateSeparator, Copy(S + DateSeparator, ii, 2 + Length(S) - ii));
         //Month is S[ii..jj[ as a LongMonthNames string
         for M := 1 to 12 do begin
            if 0 = CompareText(Copy(S, ii, jj - ii), LongMonthNames[M]) then begin
               Month := M;
               break;
            end; { then }
         end; { for }
         if Month = 0 then
            raise EConvertError.CreateFmt(SInvalidDate, [S]);
      end else if 0 = CompareText(SS, 'd') then begin
         jj := (ii - 1) + Pos(DateSeparator, Copy(S + DateSeparator, ii, 2 + Length(S) - ii));
         //Day is S[ii..jj[
         try
            Day := StrToInt(Copy(S, ii, jj - ii));
         except
            raise EConvertError.CreateFmt(SInvalidDate, [S]);
         end; { try }
      end else if 0 = CompareText(SS, 'dd') then begin
         jj := ii + 2;
         //Day is S[ii..jj[
         try
            Day := StrToInt(Copy(S, ii, jj - ii));
         except
            raise EConvertError.CreateFmt(SInvalidDate, [S]);
         end; { try }
      end else if 0 = CompareText(SS, 'ddd') then begin
         jj := (ii - 1) + Pos(DateSeparator, Copy(S + DateSeparator, ii, 2 + Length(S) - ii));
         //Day is S[ii..jj[ as a ShortDayNames string
         for M := 1 to 7 do begin
            if 0 = CompareText(Copy(S, ii, jj - ii), ShortDayNames[M]) then begin
               Day := M;
               break;
            end; { then }
         end; { for }
         if Day = 0 then
            raise EConvertError.CreateFmt(SInvalidDate, [S]);
      end else if 0 = CompareText(SS, 'dddd') then begin
         jj := (ii - 1) + Pos(DateSeparator, Copy(S + DateSeparator, ii, 2 + Length(S) - ii));
         //Day is S[ii..jj[ as a LongDayNames string
         for M := 1 to 7 do begin
            if 0 = CompareText(Copy(S, ii, jj - ii), LongDayNames[M]) then begin
               Day := M;
               break;
            end; { then }
         end; { for }
         if Day = 0 then
            raise EConvertError.CreateFmt(SInvalidDate, [S]);
      end else begin
         raise EConvertError.CreateFmt(SInvalidDateFormat, [ShortDateFormat]);
      end;
      ii := jj + 1;
      i  :=  j + 1;
   end; { while }
   try
      Result := EncodeDate(Year, Month, Day);
   except
      raise EConvertError.CreateFmt(SInvalidDate, [S]);
   end;
end; { DUTStrToDate }

function IntToRoman(Num : integer) : string;
{
   Returns the roman numeral representation of Num.
An exception is raised if Num is outside the range [1..3999].
   Larger numbers are possible but cannot be represented
in ANSI characters; 10,000, for example, is represented by an
X surmounted by a horizontal bar --e.g. "overlined X".
   An improved and corrected version of Shane Holmes' romanNum function.
}
const
   Digit = 'IXCM';
   Pivot = 'VLD';
var
   i, n : integer;
begin
   if (Num < 1) or (Num > 3999) then
      raise ERangeError.Create('IntToRoman: Argument must be in the [1..3999] range.');
   Result := '';
   for i := 1 to 4 do begin
      // i == units (1), tens (2), hundreds (3), thousands (4)
      n := Num mod 10;
      Num := Num div 10;
      //N is in [0..9]
      case n of
      0 :
         ; { Result unchanged }
      1..3 :
         Result := StringOfChar(Digit[i], n) + Result;
      4 :
         Result := Digit[i] + Pivot[i] + Result;
      5..8 :
         Result := Pivot[i] + StringOfChar(Digit[i], n - 5) + Result;
      9 :
         Result := Copy(Digit, i, 2) + Result;
      end; { case }
   end; { for }
end; { IntToRoman }

end. { DUTUtilsStrings }
