unit DUTUtilsVCL;
(*
   Daniel U. Thibault
   <D.U.Thibault@Bigfoot.com>
   12 July 1999
   Updated 3 May 2001
    1998-2001 Daniel U. Thibault.  All Rights Reserved.

   This is freeware.  No legally binding promises.  If you use
   any part of this code, you must include an acknowledgement in the
   finished code or executable (in the About box and/or the Help file).

   Various VCL utilities.
*)
interface

uses
   Classes,  { TThread }
   Controls, { TControl }
   SysUtils, { TMethod, ANSICompareText, Exception, etc. }
   ComCtrls, { TTreeView }
   Menus;    { TShortcut }

{ #################### THotKey Utilities #################### }

//Incredibly, THotKey's HotKeyToShortcut and ShortcutToHotkey functions
//are private!  We expose them in somewhat modified form here.

function HotKeyToShortcut(Value : Word) : TShortCut;
{
   Transforms a WinAPI hot key Word into a VCL TShortcut.
}

function ShortCutToHotKey(Value: TShortCut) : Word;
{
   Transforms a VCL TShortcut into a WinAPI hot key Word.
}

{ #################### TTreeView Utilities #################### }

function FindNodeByText(TreeView : TTreeView; Text : String) : TTreeNode;
{
   Finds a tree node by its Text (not case-sensitive).
   Traverses the entire TreeView, regardless of visibility.
   Returns Nil if no match.
}

{ #################### Thread Utilities #################### }

function ThreadExists(Thread : TThread) : Boolean;
{
   Returns True if the Thread exists.
}

function ThreadRunning(Thread : TThread) : Boolean;
{
   Returns True if the Thread exists and is running.
}

{ #################### Object Utilities #################### }

function HasProperty(AnObject : TObject; PropertyName : String) : Boolean;
{
   Returns True if AnObject has the property PropertyName,
   False otherwise.
}

function GetFloatProperty(Instance : TObject; PropertyName : String) : Extended;
{
   Returns Instance.Property, if Instance has that property;
   if Instance does not have that property, returns zero.
   Property must be a real type (Float, Single, Double, Comp, etc.).
}

function GetMethodProperty(Instance : TObject; PropertyName : String) : TMethod;
{
   Returns Instance.Property, if Instance has that property;
   if Instance does not have that property, returns an empty TMethod.
   Property must be a method type.
}

function GetOrdProperty(Instance : TObject; PropertyName : String) : LongInt;
{
   Returns Instance.Property, if Instance has that property;
   if Instance does not have that property, returns zero.
   Property must be an ordinal type (Boolean, Integer, etc.).
}

function GetStrProperty(Instance : TObject; PropertyName : String) : String;
{
   Returns Instance.Property, if Instance has that property;
   if Instance does not have that property, returns an empty string.
   Property must be a String type.
}

function GetVariantProperty(Instance : TObject; PropertyName : String) : Variant;
{
   Returns Instance.Property, if Instance has that property;
   if Instance does not have that property, returns Unassigned.
   Property must be a variant type.
}

procedure SetFloatProperty(Instance : TObject; PropertyName : String; Value : Extended);
{
   Sets Instance.Property to Value, if Instance has that property;
   raises an exception otherwise.
   Property must be a real type (Float, Single, Double, Comp, etc.).
}

procedure SetMethodProperty(Instance : TObject; PropertyName : String; Value : TMethod);
{
   Sets Instance.Property to Value, if Instance has that property;
   raises an exception otherwise.
   Property must be a method type.
}

procedure SetOrdProperty(Instance : TObject; PropertyName : String; Value : LongInt);
{
   Sets Instance.Property to Value, if Instance has that property;
   raises an exception otherwise.
   Property must be an ordinal type (Boolean, Integer, etc.).
}

procedure SetStrProperty(Instance : TObject; PropertyName : String; Value : String);
{
   Sets Instance.Property to Value, if Instance has that property;
   raises an exception otherwise.
   Property must be a String type.
}

procedure SetVariantProperty(Instance : TObject; PropertyName : String; Value : Variant);
{
   Sets Instance.Property to Value, if Instance has that property;
   raises an exception otherwise.
   Property must be a variant type.
}

procedure TransferOwnership(AComponent, BComponent : TComponent);
{
   Sets Ownership of all of AComponent's children (recursively) to BComponent.
   Note that AComponent can only have children if it is a TWinControl.
}

function CloneControl(AControl : TControl) : TControl;
{
   Creates a clone of an existing AControl, including all
   of its properties, events and children.
   The clone will be owned by the same TForm as AControl.
   To avoid Name conflicts, the clone's children are owned
   by the clone, not by the form.  Use TransferOwnership (q.v.)
   to restore the children's ownership once their Names are deconflicted.
   Particularly useful when loading a TPageControl with TTabSheets.

   Some controls object to this cloning method, always for the same reason:
   the cloned control has no Parent yet.  TMemo, TListBox and TComboBox
   children fail to retrieve .Items.List.  A TRichEdit child fails to
   retrieve .Lines.Strings.  A TTabControl child fails to retrieve .TabIndex.
   A THotKey child fails to retrieve .HotKey.  A TDateTimePicker child fails
   to retrieve .Date.  A TTreeView child fails to retrieve .Indent.
   A TListView child fails to retrieve .Items.Data (only if it contains
   some Items).
   A TPageControl child fails without giving details, but only if it has
   TTabSheet children.  A TAnimate child fails without giving details.
   A TToolBar child fails without giving details, but only if it has
   Images.
   TStringGrid and TDrawGrid do not fail, but they do not clone their
   Cells contents (because they are not Published properties).
   The following are "certified" as having no problem:
   [Standard]
   TPopupMenu, TLabel, TEdit, TButton, TCheckBox, TRadioButton, TScrollBar,
   TGroupBox, TRadioGroup, TPanel
   [Additional]
   TBitBtn, TSpeedButton, TMaskEdit, TImage, TShape, TBevel, TScrollBox,
   TCheckListBox, TSplitter, TStaticText, TChart
   [Win32]
   TImageList, TTrackBar, TProgressBar, TUpDown, THeaderControl, TStatusBar,
   TCoolBand

   Based on code by
   Peter Below (TeamB)  <100113.1101@compuserve.com>  12 August 1999

   Usage example:
   var Sheet : TTabSheet;
   Sheet := CloneControl(TabSheet1) as TTabSheet;
   Sheet.PageControl := PageControl1;
   Sheet.Caption := Format('TabSheet%d', [PageControl1.PageCount]);
   Sheet.Name := Sheet.Caption;
}

implementation

uses
   Windows,  { GetExitCodeThread, etc. }
   TypInfo;  { PTypeInfo, etc. }

{ #################### THotKey Utilities #################### }

type
   THKModifier  = (hkShift, hkCtrl, hkAlt, hkExt);
   THKModifiers = set of THKModifier;

function HotKeyToShortcut(Value : Word) : TShortCut;
{
   Transforms a WinAPI hot key Word into a VCL TShortcut.
}
var
   FHotKey    : Byte;
   FModifiers : THKModifiers;
begin
   Byte(FModifiers) := LoWord(HiByte(Value));
   FHotKey := LoWord(LoByte(Value));
   Result := FHotKey;
   if hkShift in FModifiers then Inc(Result, scShift);
   if hkCtrl  in FModifiers then Inc(Result, scCtrl);
   if hkAlt   in FModifiers then Inc(Result, scAlt);
end; { HotKeyToShortcut }

function ShortCutToHotKey(Value: TShortCut) : Word;
{
   Transforms a VCL TShortcut into a WinAPI hot key Word.
}
var
   FHotKey    : Byte;
   FModifiers : THKModifiers;
begin
   FHotKey := Value and not (scShift + scCtrl + scAlt);
   FModifiers := [];
   if 0 <> (Value and scShift) then Include(FModifiers, hkShift);
   if 0 <> (Value and scCtrl)  then Include(FModifiers, hkCtrl);
   if 0 <> (Value and scAlt)   then Include(FModifiers, hkAlt);
   Result := MakeWord(FHotKey, Byte(FModifiers));
end; { ShortCutToHotKey }

{ #################### TTreeView Utilities #################### }

function FindNodeByText(TreeView : TTreeView; Text : String) : TTreeNode;
{
   Finds a tree node by its Text (not case-sensitive).
   Traverses the entire TreeView, regardless of visibility.
   Returns Nil if no match.
}
var
   Node : TTreeNode;
begin
   Result := Nil;
   if not Assigned(TreeView) then Exit;
   Node := TreeView.Items.GetFirstNode;
   while Assigned(Node) do begin
      if 0 = ANSICompareText(Node.Text, Text) then begin
         Result := Node;
         break;
      end; { then }
      Node := Node.GetNext;
   end; { while }
end; { FindNodeByText }

{ #################### Thread Utilities #################### }

function ThreadExists(Thread : TThread) : Boolean;
{
   Returns True if the Thread exists.
}
var
   ExitCode : Integer;
begin
   Result := False;
   try
      {
      Although the Thread may have been Freed,
      looking up its handle does not necessarily raise an exception
      }
      Result := GetExitCodeThread(Thread.Handle, ExitCode);
   except
      //Definitely does not exist
   end;
end; { ThreadExists }

function ThreadRunning(Thread : TThread) : Boolean;
{
   Returns True if the Thread exists and is running.
}
var
   ExitCode : Integer;
begin
   Result := False;
   try
      {
      Although the Thread may have been Freed,
      looking up its handle does not necessarily raise an exception
      }
      Result := GetExitCodeThread(Thread.Handle, ExitCode);
      if Result then Result := (ExitCode = STILL_ACTIVE);
   except
      //Definitely does not exist (and therefore isn't running)
   end;
end; { ThreadRunning }

{ #################### Object Utilities #################### }

function HasProperty(AnObject : TObject; PropertyName : String) : Boolean;
{
   Returns True if AnObject has the property PropertyName,
   False otherwise.
}
var
   Info : PTypeInfo;
   Data : PTypeData;
   Size,
   i    : Integer;
   List : PPropList;
begin
   Result := False;
   Info := AnObject.ClassInfo;
   Data := GetTypeData(Info);
   Size := Data^.PropCount * SizeOf(TPropInfo);
   GetMem(List, Size);
   try
      GetPropInfos(Info, List);
      for i := 0 to (Data^.PropCount - 1) do begin
         if 0 = CompareText(List^[i]^.Name, PropertyName) then begin
            Result := True;
            Exit;
         end; { then }
      end; { for }
   finally
      FreeMem(List, Size);
   end; { try }
end; { HasProperty }

function GetFloatProperty(Instance : TObject; PropertyName : String) : Extended;
{
   Returns Instance.Property, if Instance has that property;
   if Instance does not have that property, returns zero.
   Property must be a real type (Float, Single, Double, Comp, etc.).
}
begin
   Result := 0.0;
   if not HasProperty(Instance, PropertyName) then Exit;
   Result := GetFloatProp(Instance, GetPropInfo(Instance.ClassInfo, PropertyName));
end; { GetFloatProperty }

function GetMethodProperty(Instance : TObject; PropertyName : String) : TMethod;
{
   Returns Instance.Property, if Instance has that property;
   if Instance does not have that property, returns an empty TMethod.
   Property must be a method type.
}
begin
   Result.Code := Nil;
   Result.Data := Nil;
   if not HasProperty(Instance, PropertyName) then Exit;
   Result := GetMethodProp(Instance, GetPropInfo(Instance.ClassInfo, PropertyName));
end; { GetMethodProperty }

function GetOrdProperty(Instance : TObject; PropertyName : String) : LongInt;
{
   Returns Instance.Property, if Instance has that property;
   if Instance does not have that property, returns zero.
   Property must be an ordinal type (Boolean, Integer, etc.).
}
begin
   Result := 0;
   if not HasProperty(Instance, PropertyName) then Exit;
   Result := GetOrdProp(Instance, GetPropInfo(Instance.ClassInfo, PropertyName));
end; { GetOrdProperty }

function GetStrProperty(Instance : TObject; PropertyName : String) : String;
{
   Returns Instance.Property, if Instance has that property;
   if Instance does not have that property, returns an empty string.
   Property must be a String type.
}
begin
   Result := '';
   if not HasProperty(Instance, PropertyName) then Exit;
   Result := GetStrProp(Instance, GetPropInfo(Instance.ClassInfo, PropertyName));
end; { GetStrProperty }

function GetVariantProperty(Instance : TObject; PropertyName : String) : Variant;
{
   Returns Instance.Property, if Instance has that property;
   if Instance does not have that property, returns Unassigned.
   Property must be a variant type.
}
begin
   Result := Unassigned;
   if not HasProperty(Instance, PropertyName) then Exit;
   Result := GetVariantProp(Instance, GetPropInfo(Instance.ClassInfo, PropertyName));
end; { GetVariantProperty }

procedure SetFloatProperty(Instance : TObject; PropertyName : String; Value : Extended);
{
   Sets Instance.Property to Value, if Instance has that property;
   raises an exception otherwise.
   Property must be a real type (Float, Single, Double, Comp, etc.).
}
begin
   if not HasProperty(Instance, PropertyName) then
      raise Exception.CreateFmt('DUTUtilsVCL.SetFloatProperty: %s does not have a %s real property.', [Instance.ClassName , PropertyName]);
   SetFloatProp(Instance, GetPropInfo(Instance.ClassInfo, PropertyName), Value);
end; { SetFloatProperty }

procedure SetMethodProperty(Instance : TObject; PropertyName : String; Value : TMethod);
{
   Sets Instance.Property to Value, if Instance has that property;
   raises an exception otherwise.
   Property must be a method type.
}
begin
   if not HasProperty(Instance, PropertyName) then
      raise Exception.CreateFmt('DUTUtilsVCL.SetMethodProperty: %s does not have a %s method property.', [Instance.ClassName , PropertyName]);
   SetMethodProp(Instance, GetPropInfo(Instance.ClassInfo, PropertyName), Value);
end; { SetMethodProperty }

procedure SetOrdProperty(Instance : TObject; PropertyName : String; Value : LongInt);
{
   Sets Instance.Property to Value, if Instance has that property;
   raises an exception otherwise.
   Property must be an ordinal type (Boolean, Integer, etc.).
}
begin
   if not HasProperty(Instance, PropertyName) then
      raise Exception.CreateFmt('DUTUtilsVCL.SetOrdProperty: %s does not have a %s ordinal property.', [Instance.ClassName , PropertyName]);
   SetOrdProp(Instance, GetPropInfo(Instance.ClassInfo, PropertyName), Value);
end; { SetOrdProperty }

procedure SetStrProperty(Instance : TObject; PropertyName : String; Value : String);
{
   Sets Instance.Property to Value, if Instance has that property;
   raises an exception otherwise.
   Property must be a String type.
}
begin
   if not HasProperty(Instance, PropertyName) then
      raise Exception.CreateFmt('DUTUtilsVCL.SetStrProperty: %s does not have a %s string property.', [Instance.ClassName , PropertyName]);
   SetStrProp(Instance, GetPropInfo(Instance.ClassInfo, PropertyName), Value);
end; { SetStrProperty }

procedure SetVariantProperty(Instance : TObject; PropertyName : String; Value : Variant);
{
   Sets Instance.Property to Value, if Instance has that property;
   raises an exception otherwise.
   Property must be a variant type.
}
begin
   if not HasProperty(Instance, PropertyName) then
      raise Exception.CreateFmt('DUTUtilsVCL.SetVariantProperty: %s does not have a %s variant property.', [Instance.ClassName , PropertyName]);
   SetVariantProp(Instance, GetPropInfo(Instance.ClassInfo, PropertyName), Value);
end; { SetVariantProperty }

procedure TransferOwnership(AComponent, BComponent : TComponent);
{
   Sets Ownership of all of AComponent's children (recursively) to BComponent.
   Note that AComponent can only have children if it is a TWinControl.
}
var
   i     : Integer;
   Child : TControl;
begin
   //Only TWinControls may have children
   if not (AComponent is TWinControl) then Exit;
   for i := 0 to ((AComponent as TWinControl).ControlCount - 1) do begin
      Child := (AComponent as TWinControl).Controls[i];
      if Child.Owner <> BComponent then begin
         Child.Owner.RemoveComponent(Child);
         BComponent.InsertComponent(Child);
      end; { then }
      TransferOwnership(Child, BComponent);
   end; { for }
end; { TransferOwnership }

procedure RegClasses(AComponent : TComponent);
{
   Private function used by CloneControl.
   Registers AComponent's class as well as those
   of its children (recursively) with the streaming system.
   Note that AComponent can only have children if it is a TWinControl.
}
var
   i : Integer;
begin
   Classes.RegisterClass(TPersistentClass(AComponent.ClassType));
   //Only TWinControls may have children
   if not (AComponent is TWinControl) then Exit;
   for i := 0 to ((AComponent as TWinControl).ControlCount - 1) do begin
      RegClasses((AComponent as TWinControl).Controls[i]);
   end; { for }
end; { RegClasses }

procedure CloneEvents(AControl, BControl : TControl);
{
   Private function used by CloneControl.
   It recursively copies the event handlers of AControl
   and its children to its clone BControl.
}
var
   NProps,
   i        : Integer;
   PropList : TPropList; // = array[0..16379] of PPropInfo;
begin
   //Obtain the list of BControl's method properties
   NProps := GetPropList(BControl.ClassInfo, [tkMethod], @PropList);
   for i := 0 to (NProps - 1) do begin
      //Skip read-only properties
      if Assigned(PropList[i]^.SetProc) then
         SetMethodProp(BControl, PropList[i], GetMethodProp(AControl, PropList[i]));
   end; { for }
   if not (BControl is TWinControl) then Exit;
   for i := 0 to ((BControl as TWinControl).ControlCount - 1) do begin
      CloneEvents((AControl as TWinControl).Controls[i], (BControl as TWinControl).Controls[i]);
   end; { for }
end; { CloneEvents }
(*
procedure CloneProperties(AControl, BControl : TControl);
{
   Private function used by CloneControl.
   It recursively copies the properties of AControl and
   its children to its clone BControl.
   The original need was to clone the public but un-published
   properties, but you can't tell them apart.
}
var
   NProps,
   i        : Integer;
   PropList : TPropList; // = array[0..16379] of PPropInfo;
begin
   //tkProperties = tkAny - tkMethods - [tkUnknown];
   //              (tkInteger, tkChar, tkEnumeration, tkFloat,
   //               tkString, tkSet, tkClass, tkWChar, tkLString, tkWString,
   //               tkVariant, tkArray, tkRecord, tkInterface);
   // We don't want to mess with tkClass

   //(There is apparently no way to distinguish between private, protected, public and published)
   //Obtain the list of BControl's ordinal properties
   NProps := GetPropList(BControl.ClassInfo, [tkInteger, tkChar, tkEnumeration, tkSet, tkWChar], @PropList);
   for i := 0 to (NProps - 1) do begin
      //Skip read-only properties
      if Assigned(PropList[i]^.SetProc) then
         SetOrdProp(BControl, PropList[i], GetOrdProp(AControl, PropList[i]));
   end; { for }
   //Obtain the list of BControl's floating point properties
   NProps := GetPropList(BControl.ClassInfo, [tkFloat], @PropList);
   for i := 0 to (NProps - 1) do begin
      //Skip read-only properties
      if Assigned(PropList[i]^.SetProc) then
         SetFloatProp(BControl, PropList[i], GetFloatProp(AControl, PropList[i]));
   end; { for }
   //Obtain the list of BControl's string properties
   NProps := GetPropList(BControl.ClassInfo, [tkString, tkLString, tkWString], @PropList);
   for i := 0 to (NProps - 1) do begin
      //Skip read-only properties
      if Assigned(PropList[i]^.SetProc) then
         SetStrProp(BControl, PropList[i], GetStrProp(AControl, PropList[i]));
   end; { for }
   //Obtain the list of BControl's variant properties
   NProps := GetPropList(BControl.ClassInfo, [tkVariant], @PropList);
   for i := 0 to (NProps - 1) do begin
      //Skip read-only properties
      if Assigned(PropList[i]^.SetProc) then
         SetVariantProp(BControl, PropList[i], GetVariantProp(AControl, PropList[i]));
   end; { for }
   // TTypeDate has no Case for tkArray or tkRecord
   // (tkArray, tkRecord, tkInterface);

   if not (BControl is TWinControl) then Exit;
   for i := 0 to ((BControl as TWinControl).ControlCount - 1) do begin
      CloneProperties((AControl as TWinControl).Controls[i], (BControl as TWinControl).Controls[i]);
   end; { for }
end; { CloneProperties }
*)
function CloneControl(AControl : TControl) : TControl;
{
   Creates a clone of an existing AControl, including all
   of its properties, events and children.
   The clone will be owned by the same TForm as AControl.
   To avoid Name conflicts, the clone's children are owned
   by the clone, not by the form.  Use TransferOwnership (q.v.)
   to restore the children's ownership once their Names are deconflicted.
   Particularly useful when loading a TPageControl with TTabSheets.

   Some controls object to this cloning method, always for the same reason:
   the cloned control has no Parent yet.  TMemo, TListBox and TComboBox
   children fail to retrieve .Items.List.  A TRichEdit child fails to
   retrieve .Lines.Strings.  A TTabControl child fails to retrieve .TabIndex.
   A THotKey child fails to retrieve .HotKey.  A TDateTimePicker child fails
   to retrieve .Date.  A TTreeView child fails to retrieve .Indent.
   A TListView child fails to retrieve .Items.Data (only if it contains
   some Items).
   A TPageControl child fails without giving details, but only if it has
   TTabSheet children.  A TAnimate child fails without giving details.
   A TToolBar child fails without giving details, but only if it has
   Images.
   TStringGrid and TDrawGrid do not fail, but they do not clone their
   Cells contents (because they are not Published properties).
   The following are "certified" as having no problem:
   [Standard]
   TPopupMenu, TLabel, TEdit, TButton, TCheckBox, TRadioButton, TScrollBar,
   TGroupBox, TRadioGroup, TPanel
   [Additional]
   TBitBtn, TSpeedButton, TMaskEdit, TImage, TShape, TBevel, TScrollBox,
   TCheckListBox, TSplitter, TStaticText, TChart
   [Win32]
   TImageList, TTrackBar, TProgressBar, TUpDown, THeaderControl, TStatusBar,
   TCoolBand

   Based on code by
   Peter Below (TeamB)  <100113.1101@compuserve.com>  12 August 1999

   Usage example:
   var Sheet : TTabSheet;
   Sheet := CloneControl(TabSheet1) as TTabSheet;
   Sheet.PageControl := PageControl1;
   Sheet.Caption := Format('TabSheet%d', [PageControl1.PageCount]);
   Sheet.Name := Sheet.Caption;
}
var
   ms : TMemoryStream;
begin
   //Make the template AControl the owner of all controls on it
   //(Normally, the Form Owns all its controls)
   //If this isn't done, the cloned TControl will be empty
   TransferOwnership(AControl, AControl);
   //Register the classes with the streaming system
   RegClasses(AControl);

   ms := TMemoryStream.Create;
   try { ms }
      ms.WriteComponent(AControl);
      //Rewind
      ms.Position := 0;
      //Instantiate clone
      Result := ms.ReadComponent(Nil) as TControl;
   finally
      ms.Free;
   end; { finally }

   //Restore ownership
   TransferOwnership(AControl, AControl.Owner);
// TransferOwnership(Result, AControl.Owner);
   //Recover the event handlers
   CloneEvents(AControl, Result);
   //Recover the un-published properties
// CloneProperties(AControl, Result);
end; { CloneControl }

end. { DUTUtilsVCL }
