unit DUTUtilsWinSock;
(*
   Daniel U. Thibault
   <D.U.Thibault@Bigfoot.com>
   12 August 1999
    1999 Daniel U. Thibault.  All Rights Reserved.

   This is freeware.  No legally binding promises.  If you use
   any part of this code, you must include an acknowledgement in the
   finished code or executable (in the About box and/or the Help file).

   Windows Socket utilities.
*)
interface

{ #################### TCP/IP Utilities #################### }

function Get_Local_IPAddr : String;
{
   Returns the local IP address(es).
   When there are multiple IP addresses, they are separated by pipes ('|').
}

implementation

uses
   Windows,
   SysUtils, { Exception }
   WinSock;  { WSAStartup, GetHostName, etc. }

{ #################### TCP/IP Utilities #################### }

function Get_Local_IPAddr : String;
{
   Returns the local IP address(es).
   A (32-bit) IP address is a string of the form
   ByteStr + '.' + ByteStr + '.' + ByteStr + '.' + ByteStr
   where ByteStr is the decimal representation of a single byte (0..255)
   When there are multiple IP addresses, they are separated by pipes ('|').
}
const
   sSignature =
      'Get_Local_IPAddr: ';
   sSYSNOTREADY =
      'The underlying network subsystem is not ready for network communication.';
   sVERNOTSUPPORTED =
      'The version of Windows Sockets support requested is not provided by this particular Windows Sockets implementation.';
   sEINVAL =
      'The Windows Sockets version specified by the application is not supported by this DLL.';
   sNOTINITIALISED =
      'Windows Sockets not initialised.';
   sENETDOWN =
      'The network subsystem has failed.';
   sEINPROGRESS =
      'A blocking Windows Sockets operation is in progress.';
   sHOST_NOT_FOUND =
      'Authoritative Answer Host not found.';
   sTRY_AGAIN =
      'Non-Authoritative Host not found or ServerFail.  Try again.';
   sNO_RECOVERY =
      'Nonrecoverable errors: FormErr, Refused or NotImp.';
   sNO_DATA =
      'No data record of requested type.';
   sEINTR =
      'The blocking call was interrupted.';

type
   TaPInAddr = array [0..10] of PInAddr;
   PaPInAddr = ^TaPInAddr;
var
   phe       : PHostEnt;
   pptr      : PaPInAddr;
   Buffer    : String;
   I         : Integer;
   GInitData : TWSAData;
begin
   Result := '';
//function WSAStartup(wVersionRequired: word; var WSData: TWSAData): Integer; stdcall;
   case WSAStartup($0101 { requesting version 1.1 }, GInitData) of
   0 : begin // no error
      try
         I := 64; //Initial buffer size (including terminating zero)
         while True do begin
            Buffer := StringOfChar(#0, I);
            if SOCKET_ERROR = GetHostName(PChar(Buffer), I) then begin
               case WSAGetLastError of
               WSAEFAULT : begin
                  //The buffer is too small
                  I := I*2;
               end; { WSAEFAULT }

               WSANOTINITIALISED	:
                  raise Exception.Create(sSignature + sNOTINITIALISED);
               WSAENETDOWN	:
                  raise Exception.Create(sSignature + sENETDOWN);
               WSAEINPROGRESS	:
                  raise Exception.Create(sSignature + sEINPROGRESS);
               end; { case }
            end else begin
               //Success
               break;
            end; { then }
         end; { while }

         phe := GetHostByName(PChar(Buffer));
         if phe = nil then begin
            case WSAGetLastError of
            WSANOTINITIALISED	:
               raise Exception.Create(sSignature + sNOTINITIALISED);
            WSAENETDOWN	:
               raise Exception.Create(sSignature + sENETDOWN);
            WSAHOST_NOT_FOUND :
               raise Exception.Create(sSignature + sHOST_NOT_FOUND);
            WSATRY_AGAIN :
               raise Exception.Create(sSignature + sTRY_AGAIN);
            WSANO_RECOVERY :
               raise Exception.Create(sSignature + sNO_RECOVERY);
            WSANO_DATA :
               raise Exception.Create(sSignature + sNO_DATA);
            WSAEINTR :
               raise Exception.Create(sSignature + sEINTR);
            WSAEINPROGRESS	:
               raise Exception.Create(sSignature + sEINPROGRESS);
            end; { case }
         end; { then }

         //The h_addr_list member of the THostEnt points to a list of addresses
         //More precisely, an array of pointers to address structures (unsigned longs)
         //The array is terminated by a Nil pointer
         pptr := PaPInAddr(Phe^.h_addr_list);
         I := 0;
         while pptr^[I] <> nil do begin
            //inet_ntoa converts a network address into a PChar in dotted format
            //It returns a Nil PChar in case of failure
            if Result = '' then begin
               Result :=                String(inet_ntoa(pptr^[I]^));
            end else begin
               Result := Result + '|' + String(inet_ntoa(pptr^[I]^));
            end; { else }
            Inc(I);
         end; { while }
      finally
         if SOCKET_ERROR = WSACleanup then begin
            case WSAGetLastError of
            WSANOTINITIALISED	:
               raise Exception.Create(sSignature + sNOTINITIALISED);
            WSAENETDOWN	:
               raise Exception.Create(sSignature + sENETDOWN);
            WSAEINPROGRESS	:
               raise Exception.Create(sSignature + sEINPROGRESS);
            end; { case }
         end; { then }
      end; { try }
   end; { no error }

   WSASYSNOTREADY	:
      raise Exception.Create(sSignature + sSYSNOTREADY);
   WSAVERNOTSUPPORTED :
      raise Exception.Create(sSignature + sVERNOTSUPPORTED);
   WSAEINVAL :
      raise Exception.Create(sSignature + sEINVAL);
   end; { case }
end; { Get_Local_IPAddr }

end. { DUTUtilsWinSock }
